@@ -19,7 +19,7 @@ from .enumerate import create_winrm_client
 from .util import ConnectionInfo, verify_conn_info, UnauthorizedError
 
 logging.basicConfig()
-log = logging.getLogger('zen.winrm')
+log = logging.getLogger('winrm')
 _exit_status = 0
 DEFAULT_SCHEME = 'http'
 DEFAULT_PORT = 5985
@@ -183,8 +183,19 @@ def _parse_config_file(filename, utility):
     for remote, cred_key in parser.items('remotes'):
         auth_type, username, password = creds[cred_key]
         hostname, scheme, port = _parse_remote(remote)
+        connectiontype = 'Keep-Alive'
+        keytab = ''
+        dcip = ''
         conn_info = ConnectionInfo(
-            hostname, auth_type, username, password, scheme, port)
+            hostname,
+            auth_type,
+            username,
+            password,
+            scheme,
+            port,
+            connectiontype,
+            keytab,
+            dcip,)
         try:
             verify_conn_info(conn_info)
         except Exception as e:
@@ -204,6 +215,8 @@ def _parse_args(utility):
     parser.add_argument("--authentication", "-a", default='basic',
                         choices=['basic', 'kerberos'])
     parser.add_argument("--username", "-u")
+    parser.add_argument("--dcip", "-i")
+    parser.add_argument("--keytab", "-k")
     utility.add_args(parser)
     args = parser.parse_args()
     if not args.config:
@@ -216,9 +229,10 @@ def _parse_args(utility):
         if args.remote:
             hostname, scheme, port = _parse_remote(args.remote)
             password = getpass()
+            connectiontype = 'Keep-Alive'
             args.conn_info = ConnectionInfo(
                 hostname, args.authentication, args.username, password, scheme,
-                port)
+                port, connectiontype, args.keytab, args.dcip)
             try:
                 verify_conn_info(args.conn_info)
             except Exception as e:
