@@ -20,7 +20,7 @@ from xml.etree import cElementTree as ET
 from . import constants as c
 from .util import create_etree_request_sender, get_datetime
 
-log = logging.getLogger('zen.winrm')
+log = logging.getLogger('winrm')
 _MAX_REQUESTS_PER_COMMAND = 9999
 
 
@@ -224,6 +224,51 @@ def create_long_running_command(conn_info):
     return LongRunningCommand(sender)
 
 
+@defer.inlineCallbacks
+def create_long_running_shell(conn_info):
+    results = {}
+
+    sender = create_etree_request_sender(conn_info)
+    elem = yield sender.send_request('create')
+    shell_id = _find_shell_id(elem)
+
+    results['sender'] = sender
+    results['shell_id'] = shell_id
+
+    defer.returnValue(results)
+
+
+@defer.inlineCallbacks
+def retrieve_long_running_shell(sender, shell_id, command_line):
+
+    stdout_parts = []
+    stderr_parts = []
+    exit_code = None
+
+    command_line_elem = _build_command_line_elem(command_line)
+
+    command_elem = yield sender.send_request(
+        'command', shell_id=shell_id,
+        command_line_elem=command_line_elem)
+    command_id = _find_command_id(command_elem)
+
+    for i in xrange(3):
+        receive_elem = yield sender.send_request(
+            'receive',
+            shell_id=shell_id,
+            command_id=command_id)
+        stdout_parts.extend(
+            _find_stream(receive_elem, command_id, 'stdout'))
+        stderr_parts.extend(
+            _find_stream(receive_elem, command_id, 'stderr'))
+        exit_code = _find_exit_code(receive_elem, command_id)
+
+    stdout = _stripped_lines(stdout_parts)
+    stderr = _stripped_lines(stderr_parts)
+
+    defer.returnValue(CommandResponse(stdout, stderr, exit_code))
+
+
 class Typeperf(object):
 
     def __init__(self, long_running_command):
