@@ -16,20 +16,18 @@ from datetime import datetime
 from collections import namedtuple
 from xml.etree import cElementTree as ET
 from twisted.internet import reactor, defer
-from twisted.internet.protocol import Protocol, ProcessProtocol
+from twisted.internet.protocol import Protocol
 from twisted.web.client import Agent
-from twisted.internet.ssl import CertificateOptions
+from twisted.internet.ssl import ClientContextFactory
 from twisted.web.http_headers import Headers
 from . import constants as c
 
-KERBEROS_INSTALLED = False
-try:
-    import kerberos
-    KERBEROS_INSTALLED = True
-except ImportError:
-    pass
+from .krb5 import kinit, ccname
 
-log = logging.getLogger('zen.winrm')
+# ZEN-15434 lazy import to avoid segmentation fault during install
+kerberos = None
+
+log = logging.getLogger('winrm')
 _XML_WHITESPACE_PATTERN = re.compile(r'>\s+<')
 _AGENT = None
 _MAX_PERSISTENT_PER_HOST = 200
@@ -46,8 +44,14 @@ _REQUEST_TEMPLATES = {}
 _CONTENT_TYPE = {'Content-Type': ['application/soap+xml;charset=UTF-8']}
 _MAX_KERBEROS_RETRIES = 3
 _MARKER = object()
-
-
+_ENCRYPTED_CONTENT_TYPE={"Content-Type" : ["multipart/encrypted;protocol=\"application/HTTP-Kerberos-session-encrypted\";boundary=\"Encrypted Boundary\""]}
+_BODY="""--Encrypted Boundary
+Content-Type: application/HTTP-Kerberos-session-encrypted
+OriginalContent: type=application/soap+xml;charset=UTF-8;Length={original_length}
+--Encrypted Boundary
+Content-Type: application/octet-stream
+{emsg}--Encrypted Boundary
+"""
 def _has_get_attr(obj, attr_name):
     attr_value = getattr(obj, attr_name, _MARKER)
     if attr_value is _MARKER:
@@ -58,33 +62,25 @@ def _has_get_attr(obj, attr_name):
 class MyWebClientContextFactory(object):
 
     def __init__(self):
-        self._options = CertificateOptions()
+        self._options = ClientContextFactory()
 
     def getContext(self, hostname, port):
         return self._options.getContext()
 
-
 def _get_agent():
-    global _AGENT
-    if _AGENT is None:
-        context_factory = MyWebClientContextFactory()
-        try:
-            # HTTPConnectionPool has been present since Twisted version 12.1
-            from twisted.web.client import HTTPConnectionPool
-            pool = HTTPConnectionPool(reactor, persistent=True)
-            pool.maxPersistentPerHost = _MAX_PERSISTENT_PER_HOST
-            pool.cachedConnectionTimeout = _CACHED_CONNECTION_TIMEOUT
-            _AGENT = Agent(reactor, context_factory,
-                           connectTimeout=_CONNECT_TIMEOUT, pool=pool)
-        except ImportError:
-            try:
-                # connectTimeout first showed up in Twisted version 11.1
-                _AGENT = Agent(
-                    reactor, context_factory, connectTimeout=_CONNECT_TIMEOUT)
-            except TypeError:
-                _AGENT = Agent(reactor, context_factory)
-    return _AGENT
-
+    context_factory = MyWebClientContextFactory()
+    try:
+        # HTTPConnectionPool has been present since Twisted version 12.1
+        from twisted.web.client import HTTPConnectionPool
+        pool = HTTPConnectionPool(reactor, persistent=True)
+        pool.maxPersistentPerHost = _MAX_PERSISTENT_PER_HOST
+        pool.cachedConnectionTimeout = _CACHED_CONNECTION_TIMEOUT
+        agent = Agent(reactor, context_factory,
+                       connectTimeout=_CONNECT_TIMEOUT, pool=pool)
+    except ImportError:
+        from _zenclient import ZenAgent
+        agent = ZenAgent(reactor, context_factory, persistent=True, maxConnectionsPerHostName=1)
+    return agent
 
 class _StringProducer(object):
     """
@@ -113,8 +109,9 @@ class _StringProducer(object):
     def stopProducing(self):
         pass
 
-
 def _parse_error_message(xml_str):
+    if not xml_str:
+        return ""
     elem = ET.fromstring(xml_str)
     text = elem.findtext('.//{' + c.XML_NS_SOAP_1_2 + '}Text').strip()
     detail = elem.findtext('.//{' + c.XML_NS_SOAP_1_2 + '}Detail/*/*').strip()
@@ -123,15 +120,20 @@ def _parse_error_message(xml_str):
 
 class _ErrorReader(Protocol):
 
-    def __init__(self):
+    def __init__(self, gssclient=None):
         self.d = defer.Deferred()
         self._data = []
+        self.gssclient = gssclient
 
     def dataReceived(self, data):
         self._data.append(data)
 
     def connectionLost(self, reason):
-        message = _parse_error_message(''.join(self._data))
+        if self.gssclient:
+            body = self.gssclient.decrypt_body(''.join(self._data))
+        else:
+            body = ''.join(self._data)
+        message = _parse_error_message(body)
         self.d.callback(message)
 
 
@@ -139,6 +141,10 @@ class RequestError(Exception):
     pass
 
 
+class ForbiddenError(RequestError):
+    pass
+
+
 class UnauthorizedError(RequestError):
     pass
 
@@ -159,46 +165,6 @@ def _get_basic_auth_header(conn_info):
     return 'Basic {0}'.format(base64.encodestring(authstr).strip())
 
 
-class KinitProcessProtocol(ProcessProtocol):
-
-    def __init__(self, password):
-        self._password = password
-        self.d = defer.Deferred()
-        self._data = ''
-
-    def outReceived(self, data):
-        log.debug("kinit wrote to stdout: {0}".format(data))
-        self._data += data
-        if 'Password for' in self._data and ':' in self._data:
-            log.debug("sending password")
-            self.transport.write('{0}\n'.format(self._password))
-            self._data = ''
-
-    def errReceived(self, data):
-        log.debug("kinit wrote to stdin: {0}".format(data))
-
-    def processExited(self, reason):
-        if reason.value.exitCode != 0:
-            log.debug("kinit failed (exit code {0}): {1} {2}".format(
-                reason.value.exitCode,
-                reason,
-                reason.value))
-
-    def processEnded(self, reason):
-        self.d.callback(None)
-
-
-@defer.inlineCallbacks
-def kinit(username, password):
-    kinit = '/usr/bin/kinit'
-    userid, realm = username.split('@')
-    args = [kinit, '{0}@{1}'.format(userid, realm.upper())]
-    log.debug('spawing kinit process: {0}'.format(args))
-    protocol = KinitProcessProtocol(password)
-    reactor.spawnProcess(protocol, kinit, args)
-    yield protocol.d
-
-
 class AuthGSSClient(object):
     """
     The Generic Security Services (GSS) API allows Kerberos implementations to
@@ -211,15 +177,24 @@ class AuthGSSClient(object):
         1  : GSSAPI step complete, or function return OK
     """
 
-    def __init__(self, service, username, password):
+    def __init__(self, service, username, password, dcip):
         """
         @param service: a string containing the service principal in the form
             'type@fqdn' (e.g. 'imap@mail.apple.com').
         """
+        # ZEN-15434 Lazy import.  import causes segmentation fault because of
+        # differing versions of the kerberos.so file.  Only import here
+        global kerberos
+        if not kerberos:
+            import kerberos
         self._service = service
         self._username = username
         self._password = password
-        result_code, self._context = kerberos.authGSSClientInit(service)
+        self._dcip = dcip
+        gssflags = kerberos.GSS_C_CONF_FLAG|kerberos.GSS_C_MUTUAL_FLAG|kerberos.GSS_C_SEQUENCE_FLAG|kerberos.GSS_C_INTEG_FLAG
+
+        os.environ['KRB5CCNAME'] = ccname(username)
+        result_code, self._context = kerberos.authGSSClientInit(service,gssflags=gssflags)
         if result_code != kerberos.AUTH_GSS_COMPLETE:
             raise Exception('kerberos authGSSClientInit failed')
 
@@ -241,7 +216,7 @@ class AuthGSSClient(object):
         return kerberos.authGSSClientStep(self._context, challenge)
 
     @defer.inlineCallbacks
-    def get_base64_client_data(self):
+    def get_base64_client_data(self, challenge=''):
         """
         @return: a string containing the base64-encoded client data to be sent
             to the server.
@@ -249,14 +224,16 @@ class AuthGSSClient(object):
         result_code = None
         for i in xrange(_MAX_KERBEROS_RETRIES):
             try:
-                result_code = self._step()
+                result_code = self._step(challenge)
                 break
             except kerberos.GSSError as e:
                 msg = e.args[1][0]
-                if msg == 'Cannot determine realm for numeric host address':
-                    raise
+                if msg == 'Cannot determine realm for numeric host address' or \
+                   msg == 'Server not found in Kerberos database':
+                       raise Exception(msg)
                 log.debug('{0}. Calling kinit.'.format(msg))
-                yield kinit(self._username, self._password)
+                yield kinit(self._username, self._password, self._dcip)
+
         if result_code != kerberos.AUTH_GSS_CONTINUE:
             raise Exception('kerberos authGSSClientStep failed ({0}).'
                             .format(result_code))
@@ -278,26 +255,75 @@ class AuthGSSClient(object):
                             .format(result_code, challenge))
         return kerberos.authGSSClientUserName(self._context)
 
+    def encrypt_body(self, body):
+        # get original length of body. wrap will encrypt in place
+        orig_len = len(body)
+        # encode before sending to wrap func
+        ebody = base64.b64encode(body)
+        # wrap it up
+        try:
+            rc,pad_len = kerberos.authGSSClientWrapIov(self._context,ebody,1)
+            if rc is not kerberos.AUTH_GSS_COMPLETE:
+                log.debug("Unable to encrypt message body")
+                return
+        except AttributeError:
+            # must be on centos 5, encryption not possible
+            return body
+
+        # get wrapped request which is in b64 encoding
+        ewrap = kerberos.authGSSClientResponse(self._context)
+        # decode wrapped request
+        payload = bytes(base64.b64decode(ewrap))
+        # add carriage returns to body
+        body = _BODY.replace('\n','\r\n')
+        body = bytes(body.format(original_length=orig_len+pad_len,emsg=payload))
+        return body
+
+    def decrypt_body(self, body):
+        try:
+            b_start = body.index("Content-Type: application/octet-stream") + \
+                      len("Content-Type: application/octet-stream\r\n")
+        except ValueError:
+            # Unencrypted data, return body
+            return body
+        b_end = body.index("--Encrypted Boundary",b_start)
+        ebody = body[b_start:b_end]
+        ebody = base64.b64encode(ebody)
+        rc = kerberos.authGSSClientUnwrapIov(self._context, ebody)
+        if rc is not kerberos.AUTH_GSS_COMPLETE:
+            log.debug("Unable to decrypt message body")
+            return
+        ewrap = kerberos.authGSSClientResponse(self._context)
+        body = base64.b64decode(ewrap)
+        return body
 
 @defer.inlineCallbacks
-def _authenticate_with_kerberos(conn_info, url):
-    if not KERBEROS_INSTALLED:
-        raise Exception('You must run "easy_install kerberos".')
+def _authenticate_with_kerberos(conn_info, url, agent, gss_client=None):
     service = '{0}@{1}'.format(conn_info.scheme.upper(), conn_info.hostname)
-    gss_client = AuthGSSClient(service, conn_info.username, conn_info.password)
+    if gss_client is None:
+        gss_client = AuthGSSClient(
+            service,
+            conn_info.username,
+            conn_info.password,
+            conn_info.dcip)
+
     base64_client_data = yield gss_client.get_base64_client_data()
     auth = 'Kerberos {0}'.format(base64_client_data)
     k_headers = Headers(_CONTENT_TYPE)
     k_headers.addRawHeader('Authorization', auth)
     k_headers.addRawHeader('Content-Length', '0')
-    response = yield _get_agent().request('POST', url, k_headers, None)
+    response = yield agent.request('POST', url, k_headers, None)
     if response.code == httplib.UNAUTHORIZED:
         raise UnauthorizedError(
             "HTTP Unauthorized received on initial kerberos request.")
+    elif response.code == httplib.FORBIDDEN:
+        raise ForbiddenError(
+            "Forbidden. Check WinRM port and version.")
     elif response.code != httplib.OK:
         proto = _StringProtocol()
         response.deliverBody(proto)
         xml_str = yield proto.d
+        xml_str = gss_client.decrypt_body(gss_client._context, xml_str)
         raise Exception(
             "status code {0} received on initial kerberos request {1}"
             .format(response.code, xml_str))
@@ -314,31 +340,25 @@ def _authenticate_with_kerberos(conn_info, url):
     k_username = gss_client.get_username(auth_details)
     log.debug('kerberos auth successful for user: {0} / {1} '
               .format(conn_info.username, k_username))
-
-
-@defer.inlineCallbacks
-def _get_url_and_headers(conn_info):
-    url = "{c.scheme}://{c.hostname}:{c.port}/wsman".format(c=conn_info)
-    headers = Headers(_CONTENT_TYPE)
-    if conn_info.auth_type == 'basic':
-        headers.addRawHeader(
-            'Authorization', _get_basic_auth_header(conn_info))
-    elif conn_info.auth_type == 'kerberos':
-        yield _authenticate_with_kerberos(conn_info, url)
-    else:
-        raise Exception('unknown auth type: {0}'.format(conn_info.auth_type))
-    defer.returnValue((url, headers))
-
+    defer.returnValue(gss_client)
 
 ConnectionInfo = namedtuple(
-    'ConnectionInfo',
-    ['hostname', 'auth_type', 'username', 'password', 'scheme', 'port'])
+    'ConnectionInfo', [
+        'hostname',
+        'auth_type',
+        'username',
+        'password',
+        'scheme',
+        'port',
+        'connectiontype',
+        'keytab',
+        'dcip'])
 
 
 def verify_hostname(conn_info):
     has_hostname, hostname = _has_get_attr(conn_info, 'hostname')
     if not has_hostname or not hostname:
-        raise Exception("hostname missing")
+        raise Exception("hostname is not resolvable")
 
 
 def verify_auth_type(conn_info):
@@ -362,9 +382,9 @@ def verify_password(conn_info):
 
 def verify_scheme(conn_info):
     has_scheme, scheme = _has_get_attr(conn_info, 'scheme')
-    if not has_scheme or scheme != 'http':
+    if not has_scheme or scheme not in ['http', 'https']:
         raise Exception(
-            "scheme must be http (https is not implemented yet): {0}"
+            "scheme must be http or https: {0}"
             .format(scheme))
 
 
@@ -374,6 +394,12 @@ def verify_port(conn_info):
         raise Exception("illegal value for port: {0}".format(port))
 
 
+def verify_connectiontype(conn_info):
+    has_connectiontype, connectiontype = _has_get_attr(conn_info, 'connectiontype')
+    if not has_connectiontype or not connectiontype:
+        raise Exception("connectiontype missing")
+
+
 def verify_conn_info(conn_info):
     verify_hostname(conn_info)
     verify_auth_type(conn_info)
@@ -381,6 +407,7 @@ def verify_conn_info(conn_info):
     verify_password(conn_info)
     verify_scheme(conn_info)
     verify_port(conn_info)
+    verify_connectiontype(conn_info)
 
 
 class RequestSender(object):
@@ -390,15 +417,43 @@ class RequestSender(object):
         self._conn_info = conn_info
         self._url = None
         self._headers = None
+        self.gssclient = None
+        self.agent = _get_agent()
+        self.authorized = False
+
+    @defer.inlineCallbacks
+    def _get_url_and_headers(self):
+        url = "{c.scheme}://{c.hostname}:{c.port}/wsman".format(c=self._conn_info)
+        if self._conn_info.auth_type == 'basic':
+            headers = Headers(_CONTENT_TYPE)
+            headers.addRawHeader('Connection', self._conn_info.connectiontype)
+            if not self.authorized:
+                headers.addRawHeader(
+                    'Authorization', _get_basic_auth_header(self._conn_info))
+                self.authorized = True
+        elif self.is_kerberos():
+            headers = Headers(_ENCRYPTED_CONTENT_TYPE)
+            headers.addRawHeader('Connection', self._conn_info.connectiontype)
+            if self.gssclient is None:
+                self.gssclient = yield _authenticate_with_kerberos(self._conn_info, url, self.agent)
+        else:
+            raise Exception('unknown auth type: {0}'.format(self._conn_info.auth_type))
+        defer.returnValue((url, headers))
 
     @defer.inlineCallbacks
     def _set_url_and_headers(self):
-        self._url, self._headers = yield _get_url_and_headers(self._conn_info)
+        self._url, self._headers = yield self._get_url_and_headers()
 
     @property
     def hostname(self):
         return self._conn_info.hostname
 
+    def is_kerberos(self):
+        return self._conn_info.auth_type == 'kerberos'
+
+    def decrypt_body(self, body):
+        return self.gssclient.decrypt_body(body)
+
     @defer.inlineCallbacks
     def send_request(self, request_template_name, **kwargs):
         log.debug('sending request: {0} {1}'.format(
@@ -406,17 +461,42 @@ class RequestSender(object):
         if not self._url or self._conn_info.auth_type == 'kerberos':
             yield self._set_url_and_headers()
         request = _get_request_template(request_template_name).format(**kwargs)
-        # log.debug(request)
-        body_producer = _StringProducer(request)
-        response = yield _get_agent().request(
-            'POST', self._url, self._headers, body_producer)
+        if self.is_kerberos():
+            encrypted_request = self.gssclient.encrypt_body(request)
+            if not encrypted_request.startswith("--Encrypted Boundary"):
+                self.headers = Headers(_CONTENT_TYPE)
+                self.headers.addRawHeader('Connection', self._conn_info.connectiontype)
+            body_producer = _StringProducer(encrypted_request)
+        else:
+            body_producer = _StringProducer(request)
+        try:
+            response = yield self.agent.request(
+                'POST', self._url, self._headers, body_producer)
+        except Exception as e:
+            raise e
         log.debug('received response {0} {1}'.format(
             response.code, request_template_name))
         if response.code == httplib.UNAUTHORIZED:
-            raise UnauthorizedError(
-                "unauthorized, check username and password.")
+            # check to see if we need to re-authorize due to lost connection
+            if self.gssclient is not None:
+                self.agent = _get_agent()
+                yield _authenticate_with_kerberos(self._conn_info, self._url, self.agent, self.gssclient)
+                try:
+                    response = yield self.agent.request(
+                        'POST', self._url, self._headers, body_producer)
+                except Exception as e:
+                    raise e
+            if response.code == httplib.UNAUTHORIZED:
+                raise UnauthorizedError(
+                    "Unauthorized: Check username and password")
+        if response.code == httplib.FORBIDDEN:
+            raise ForbiddenError(
+                "Forbidden: Check WinRM port and version")
         elif response.code != httplib.OK:
-            reader = _ErrorReader()
+            if self.is_kerberos():
+                reader = _ErrorReader(self.gssclient)
+            else:
+                reader = _ErrorReader()
             response.deliverBody(reader)
             message = yield reader.d
             raise RequestError("HTTP status: {0}. {1}".format(
@@ -449,7 +529,18 @@ class EtreeRequestSender(object):
             request_template_name, **kwargs)
         proto = _StringProtocol()
         resp.deliverBody(proto)
-        xml_str = yield proto.d
+        body = yield proto.d
+        if self._sender.is_kerberos():
+            xml_str = self._sender.gssclient.decrypt_body(body)
+        else:
+            xml_str = yield body
+        if log.isEnabledFor(logging.DEBUG):
+            try:
+                import xml.dom.minidom
+                xml = xml.dom.minidom.parseString(xml_str)
+                log.debug(xml.toprettyxml())
+            except:
+                log.debug('Could not prettify response XML: "{0}"'.format(xml_str))
         defer.returnValue(ET.fromstring(xml_str))
 
 
@@ -458,18 +549,22 @@ def create_etree_request_sender(conn_info):
     return EtreeRequestSender(sender)
 
 
+TZOFFSET_PATTERN = re.compile(r'[-+]\d+:\d\d$')
+
+
 def get_datetime(text):
     """
     Parse the date from a WinRM response and return a datetime object.
     """
-    if text.endswith('Z'):
-        if '.' in text:
+    text2 = TZOFFSET_PATTERN.sub('Z', text)
+    if text2.endswith('Z'):
+        if '.' in text2:
             format = "%Y-%m-%dT%H:%M:%S.%fZ"
-            date_string = _NANOSECONDS_PATTERN.sub(r'.\g<1>', text)
+            date_string = _NANOSECONDS_PATTERN.sub(r'.\g<1>', text2)
         else:
             format = "%Y-%m-%dT%H:%M:%SZ"
-            date_string = text
+            date_string = text2
     else:
         format = '%m/%d/%Y %H:%M:%S.%f'
-        date_string = text
+        date_string = text2
     return datetime.strptime(date_string, format)
